#!/usr/bin/python -u
from subprocess import PIPE, Popen
from copy import deepcopy
from random import randint
import itertools

def unique(x):
	return list(set(x))

cnf_true = 1
cnf_false = -1
cnf_new_var_counter = 1
cnf_or_clauses = [[1]]
cnf_solution = []
cnf_conflicts = 0
cnf_time = 0
cnf_memory = 0

def cnf_reset():
	global cnf_new_var_counter
	global cnf_or_clauses
	cnf_new_var_counter = 1
	cnf_or_clauses = [[1]]

def cnf_generate():
	result = ""
	num_vars = cnf_new_var_counter
	num_clauses = len(cnf_or_clauses)
	result += "p cnf " + str(num_vars) + " " + str(num_clauses) + "\n"
	print result
	for a in cnf_or_clauses:
		result += ' '.join(str(x) for x in a) + ' 0\n'
	return result

def cnf_solve():
	global cnf_solution
	global cnf_conflicts
	global cnf_time
	global cnf_memory
	print "Starting cryptominisat3"
	p = Popen(["cryptominisat3"], stdin=PIPE, stdout=PIPE, bufsize=1)
	print "Writing problem"
	p.stdin.write(cnf_generate())
	p.stdin.close()
	print "Solving"
	for line in p.stdout:
		if line[0] == 'v':
			cnf_solution = [int(x) for x in line[2:].split(' ')]
		elif line.startswith('s UNSATISFIABLE'):
			print line
			cnf_solution = ['X' for x in range(cnf_new_var_counter+1)]
		else:
			print line[:-1]
			if line.startswith('c Conflicts'):
				cnf_conflicts = float(line[29:41])
			if line.startswith('c Total time'):
				cnf_time = float(line[29:41])
			if line.startswith('c Mem used'):
				cnf_memory = float(line[29:41])

def cnf_new_var():
	global cnf_new_var_counter
	cnf_new_var_counter += 1
	return cnf_new_var_counter

def cnf_or_clause(*args):
	global cnf_or_clauses
	args = unique(args)
	if cnf_false in args:
		args.remove(cnf_false)
	if cnf_true in args:
		return
	for x in args:
		if -x in args:
			return # always true
	if len(args) == 0:
		return
	cnf_or_clauses.append(args)

def cnf_true_clause(*args):
	for x in args:
		cnf_or_clause(x)

def cnf_false_clause(*args):
	for x in args:
		cnf_or_clause(cir_not(x))

def cnf_equals2_clause(a, b):
	cnf_or_clause(cir_not(a), b)
	cnf_or_clause(a, cir_not(b))

def cnf_equals_clause(*args):
	for b in args:
		cnf_equals2_clause(args[0], b)

def cnf_value(a):
	return cnf_solution[abs(a) - 1] == a

def cnf_is_constant(*args):
	"""Returns true if all the arguments are constant"""
	for b in args:
		if abs(b) != cnf_true:
			return False
	return True

def cnf_some_constant(*args):
	"""Returns true if some arguments are constant"""
	for b in args:
		if abs(b) == cnf_true:
			return True
	return False

def cnf_count_constant(*args):
	"""Returns the number of constant arguments"""
	count = 0
	for b in args:
		if abs(b) == cnf_true:
			count += 1
	return count

def cnf_is_special(*args):
	"""Returns true if some arguments are constant or redundant"""
	if cnf_some_constant(*args):
		return True
	args = [abs(x) for x in args]
	if len(unique(args)) < len(args):
		print args
		return True
	return False

#
# Circuits
#

def cir_not(a):
	return -a

def cir_and(*args):
	args = unique(args)
	if cnf_false in args:
		return cnf_false
	if cnf_true in args:
		args.remove(cnf_true)
	for x in args:
		if -x in args:
			return cnf_false
	if len(args) == 0:
		return cnf_true
	if len(args) == 1:
		return args[0]
	out = cnf_new_var()
	cnf_or_clause(*([out] + [cir_not(x) for x in args]))
	for x in args:
		cnf_or_clause(cir_not(out), x)
	return out

def cir_nand(*args):
	return cir_not(cir_and(args))

def cir_or(*args):
	return cir_not(cir_and(*[cir_not(x) for x in args]))

def cir_nor(*args):
	return cir_not(cir_or(args))

def cir_xor(*args):
	variables =unique(args)
	newlist = []
	for x in variables: # a^a^a = a
		if args.count(x) % 2 == 1:
			newlist.append(x)
	args = newlist
	outs = 1
	if cnf_true in args:
		args.remove(cnf_true)
		outs = -outs
	if cnf_false in args:
		args.remove(cnf_false)
	for x in args:
		if -x in args:
			args.remove(x)
			args.remove(-x)
			outs = -outs
	if len(args) == 0:
		return outs * cnf_false
	if len(args) == 1:
		return outs * args[0]	
	out = cnf_new_var()
	v = [out] + list(args)
	for l in range(1, len(v) + 1, 2):
		for n in itertools.combinations(v, l):
			cnf_or_clause(*[(cir_not(x) if x in n else x) for x in v])
	return outs * out

def cir_xnor(*args):
	return cir_not(cir_xor(args))

def cir_full_adder(a, b, c):
	if cnf_is_special(a, b, c):
		# Tseitin mode in special cases, gives constant propagation
		half = cir_xor(a, b)
		out = cir_xor(half, c)
		if c == cnf_false:
			carry = cir_and(a, b)
		else:
			carry = cir_or(cir_and(a, b), cir_and(half, c))
	else:
		out = cnf_new_var()
		carry = cnf_new_var()
		cnf_or_clause(out, a, b, -c)
		cnf_or_clause(out, a, -b, c)
		cnf_or_clause(out, -a, b, c)
		cnf_or_clause(out, -a, -b, -c)
		cnf_or_clause(-out, a, b, c)
		cnf_or_clause(-out, a, -b, -c)
		cnf_or_clause(-out, -a, b, -c)
		cnf_or_clause(-out, -a, -b, c)
		cnf_or_clause(carry, a, -b, -c)
		cnf_or_clause(carry, -a, b, -c)
		cnf_or_clause(carry, -a, -b, c)
		cnf_or_clause(carry, -a, -b, -c)
		cnf_or_clause(-carry, a, b, c)
		cnf_or_clause(-carry, a, b, -c)
		cnf_or_clause(-carry, a, -b, c)
		cnf_or_clause(-carry, -a, b, c)
	return out, carry

#
# Bitvectors
#

def vec_new(bits):
	return [cnf_new_var() for i in range(bits)]

def vec_true(bits):
	return [cnf_true for i in range(bits)]

def vec_false(bits):
	return [cnf_false for i in range(bits)]

def vec_const(bits, c):
	out = []
	for i in range(bits):
		if c & 1:
			out.append(cnf_true)
		else:
			out.append(cnf_false)
		c >>= 1
	return out

def vec_random(bits):
	out = []
	for i in range(bits):
		if randint(0, 1) == 1:
			out.append(cnf_true)
		else:
			out.append(cnf_false)
	return out

def vec_from_bytes(byteArray):
	out = []
	for b in bytearray(byteArray):
		out += vec_const(8, b)
	return out

def vec_from_hex(string):
	out = []
	for c in string:
		if c == 'x':
			out += vec_random(4)
		elif c == '_':
			out += vec_new(4)
		else:
			out += vec_const(4, int(c,16))
	return out

def vec_not(a):
	return [cir_not(b) for b in a]

def vec_and(*args):
	return [cir_and(*p) for p in zip(*args)]

def vec_nand(*args):
	return [cir_nand(*p) for p in zip(*args)]

def vec_or(*args):
	return [cir_or(*p) for p in zip(*args)]

def vec_nor(*args):
	return [cir_nor(*p) for p in zip(*args)]

def vec_xor(*args):
	return [cir_xor(*p) for p in zip(*args)]

def vec_xnor(*args):
	return [cir_xnor(*p) for p in zip(*args)]

def vec_rotate_left(n, b):
	return n[-b:] + n[:-b]

def vec_rotate_right(n, b):
	return n[b:] + n[:b]

def vec_shift_left(n, b):
	return vec_false(b) + n[:-b]

def vec_shift_right(n, b):
	return list(n[b:]) + vec_false(b)

def vec_bswap16(n):
	return n[8:16] + n[0:8]

def vec_bswap32(n):
	return vec_bswap16(n[16:32]) + vec_bswap16(n[0:16])

def vec_bswap64(n):
	return vec_bswap32(n[32:64]) + vec_bswap32(n[0:32])

def vec_equals(*args):
	for p in zip(*args):
		cnf_equals_clause(*p)

def vec_value(a):
	return [cnf_value(b) for b in a]

def vec_to_bytes(a):
	out = bytearray()
	for i in range(0, len(a), 8):
		out.append(int_value(a[i:i+8]))
	return out

def vec_to_hex(a):
	return "".join("%02x" % b for b in vec_to_bytes(a))

#
# Integers
#

def int_negate(n):
	return int_add(vec_not(n), vec_const(len(n), 1))

def int_sub(a, b):
	return int_add(a, vec_negate(b))

def int_add2(a, b):
	out = []
	carry = cnf_false
	pairs = zip(a, b)
	for (a, b) in pairs[:-1]:
		res, carry = cir_full_adder(a, b, carry)
		out.append(res)
	out.append(cir_xor(pairs[-1][0], pairs[-1][1], carry))
	return out

def int_add(*args):
	return reduce(int_add2, args)

def int_value(n):
	r = 0
	for b in reversed(vec_value(n)):
		r <<= 1
		if b:
			r += 1
	return r

#
# SHA256
#

sha_h = [vec_const(32, c) for c in [
	0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a, 0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19
]]

sha_k = [vec_const(32, c) for c in [
	0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
	0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
	0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
	0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
	0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
	0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
	0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
	0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
]]

def cir_choose(e, f, g):
	if cnf_is_constant(e) or cnf_is_constant(f, g):
		# Tseitin mode gives constant propagation
		# Variable is eliminated if e or (f, g) is constant
		return cir_xor(cir_and(e, f), cir_and(cir_not(e), g))
	else:
		# Optimized CNF
		out = cnf_new_var()
		cnf_or_clause(out, cir_not(e), cir_not(f))
		cnf_or_clause(out, e, cir_not(g))
		cnf_or_clause(cir_not(out), cir_not(e), f)
		cnf_or_clause(cir_not(out), e, g)
		return out

def sha_choose(e, f, g):
	return [cir_choose(*p) for p in zip(e, f, g)]

def cir_majority(a, b, c):
	if cnf_count_constant(a, b, c) >= 2:
		# Tseitin mode gives constant propagation
		# Propagation eliminates all variables for two or three constants
		return cir_xor(cir_and(a, b), cir_and(a, c), cir_and(b, c))
	else:
		# Optimized CNF
		out = cnf_new_var()
		cnf_or_clause(out, cir_not(a), cir_not(b))
		cnf_or_clause(out, cir_not(a), cir_not(c))
		cnf_or_clause(out, cir_not(b), cir_not(c))
		cnf_or_clause(cir_not(out), a, b)
		cnf_or_clause(cir_not(out), a, c)
		cnf_or_clause(cir_not(out), b, c)
		return out

def sha_majority(a, b, c):
	return [cir_majority(*p) for p in zip(a, b, c)]

def sha_sigma(a, x, y, z):
	return vec_xor(
		vec_rotate_right(a, x),
		vec_rotate_right(a, y),
		vec_shift_right(a, z))

def sha_Sigma(a, x, y, z):
	return vec_xor(
		vec_rotate_right(a, x),
		vec_rotate_right(a, y),
		vec_rotate_right(a, z))

def sha_pad(m):
	l = len(m)
	out = deepcopy(m)
	out += vec_const(8, 128)
	out += vec_false(440 - (l % 512))
	out += vec_bswap64(vec_const(64, l))
	return out

def sha_w(m, rounds):
	# m is a 512 bit message
	assert len(m) == 512
	w = [vec_bswap32(x) for x in zip(*[iter(m)]*32)]
	for i in range(16, rounds):
		w.append(int_add(
			w[i-16], w[i-7],
			sha_sigma(w[i-15], 7, 18, 3),
			sha_sigma(w[i-2], 17, 19, 10)))
	return w

def sha_round(a, b, c, d, e, f, g, h, k, w):
	temp1 = int_add(h, sha_Sigma(e, 6, 11, 25), sha_choose(e, f, g), k, w)
	temp2 = int_add(sha_Sigma(a, 2, 13, 22), sha_majority(a, b, c))
	h = g
	g = f
	f = e
	e = int_add(d, temp1)
	d = c
	c = b
	b = a
	a = int_add(temp1, temp2)
	return a, b, c, d, e, f, g, h

def sha256(m, rounds=64):
	m = sha_pad(m)
	H = deepcopy(sha_h)
	for block in zip(*[iter(m)]*512):
		w = sha_w(block, rounds)
		a, b, c, d, e, f, g, h = H
		for i in range(rounds):
			a, b, c, d, e, f, g, h = sha_round(a, b, c, d, e, f, g, h, sha_k[i], w[i])
		H[0] = int_add(a, H[0])
		H[1] = int_add(b, H[1])
		H[2] = int_add(c, H[2])
		H[3] = int_add(d, H[3])
		H[4] = int_add(e, H[4])
		H[5] = int_add(f, H[5])
		H[6] = int_add(g, H[6])
		H[7] = int_add(h, H[7])
	return reduce(lambda a, b: a+b, map(vec_bswap32, H))


results = []

#m = vec_from_hex(
#	"01000000" +
#	"81cd02ab7e569e8bcd9317e2fe99f2de44d49ab2b8851ba4a308000000000000" +
#	"e320b6c2fffc8d750423db8b1eb942ae710e951ed797f7affc8892b0f1fc122b" +
#	"c7f5d74d" +
#	"f2b9441a" +
#	"42a14695")
#vec_equals(k, vec_from_hex('1dbd981fe6985776b644b173a4d0385ddc1aa2a829688d1e0000000000000000'))

def test(n):
	global results
	
	rounds = 32
	
	repeats = 0.0
	total_time = 0.0
	total_mem = 0.0
	total_vars = 0.0
	total_clauses = 0.0
	total_conflicts = 0.0
	while (repeats < 3 or total_time < 60) and repeats < 10:
		cnf_reset()
		m = vec_from_hex(
			"01000000" +
			"xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx0000000000000000" +
			"xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx" +
			"533bff42" +
			"1900db99" +
			"________")
		h = sha256(m, rounds)
		k = sha256(h, rounds)
		#vec_equals(k[-n:], vec_const(n, 0))
		
		cnf_solve()
		
		print vec_to_hex(k)
		repeats += 1
		total_time += cnf_time
		total_mem += cnf_memory
		total_vars += float(cnf_new_var_counter)
		total_clauses += float(len(cnf_or_clauses))
		total_conflicts += cnf_conflicts
		print results
		print "\n".join(["| " + " | ".join([str(int(round(y))) for y in x]) for x in results])
		print repeats, total_time, total_mem, total_vars, total_clauses, total_conflicts
	
	total_time /= repeats
	total_mem /= repeats
	total_vars /= repeats
	total_clauses /= repeats
	total_conflicts /= repeats
	hash_rate = (2.0 ** n) / total_time
	results.append([n, total_time, total_mem, total_vars, total_clauses, total_conflicts, hash_rate])
	print results
	print "\n".join(["| " + " | ".join([str(int(round(y))) for y in x]) for x in results])

#for i in range(0, 68):
#	test(i)

#for i in range(16,64):
#	test(i)

cnf_reset()
m = vec_from_hex("1337__")
h = sha256(m, 64)
vec_equals(h[-16:], vec_from_hex("1337"))

print cnf_generate()
cnf_solve()

print vec_to_hex(m), vec_to_hex(h)





